@@ -1,4 +1,5 @@
 #include "cache.h"
+#include "blob.h"
 
 static char *create_temp_file(unsigned char *sha1)
 {
@@ -9,7 +10,7 @@ static char *create_temp_file(unsigned char *sha1)
 	int fd;
 
 	buf = read_sha1_file(sha1, type, &size);
-	if (!buf || strcmp(type, "blob"))
+	if (!buf || strcmp(type, blob_type))
 		die("unable to read blob object %s", sha1_to_hex(sha1));
 
 	strcpy(path, ".merge_file_XXXXXX");
@@ -26,10 +27,13 @@ int main(int argc, char **argv)
 {
 	unsigned char sha1[20];
 
-	if (argc != 2 || get_sha1(argv[1], sha1))
+	if (argc != 2)
 		usage("git-unpack-file <sha1>");
+	if (get_sha1(argv[1], sha1))
+		die("Not a valid object name %s", argv[1]);
 
 	setup_git_directory();
+	git_config(git_default_config);
 
 	puts(create_temp_file(sha1));
 	return 0;
